#include "xsmul.h"

/*
 *  Scalar times a vector
 */

void dsmul (int n, double alpha, double *x,  double *y)
{
  while (n--)
    *y++ = alpha * (*x++);
}


void  dsmul_disjoint(int N, double alpha, double *x, double *y){

#pragma disjoint(*x, *y)

   for (int i = 0; i < N; i++)
      y[i] = alpha*x[i];

}

void  dsmul_disjoint_double_hummer(int N, double alpha, double *x, double *y){

  int i, j;
  double _Complex x1, y1, x2, y2, x3, y3, x4, y4, x5, y5, x6, y6, x7, y7, x8, y8;

  double _Complex a;
  a = __cmplx(alpha,alpha);

  for (i=0; i < (N-(N%16)); i += 16){

        x1 = __lfpd( &x[i] );
        x2 = __lfpd( &x[i+2] );
        x3 = __lfpd( &x[i+4] );
        x4 = __lfpd( &x[i+6] );
        x5 = __lfpd( &x[i+8] );
        x6 = __lfpd( &x[i+10] );
        x7 = __lfpd( &x[i+12] );
        x8 = __lfpd( &x[i+14] );


        y1 = __fpmul(a,x1);
        y2 = __fpmul(a,x2);
        y3 = __fpmul(a,x3);
        y4 = __fpmul(a,x4);
        y5 = __fpmul(a,x5);
        y6 = __fpmul(a,x6); 
        y7 = __fpmul(a,x7);
        y8 = __fpmul(a,x8);


        __stfpd( &y[i], y1 );
        __stfpd( &y[i+2], y2 );
        __stfpd( &y[i+4], y3 );
        __stfpd( &y[i+6], y4 );
        __stfpd( &y[i+8], y5 );
        __stfpd( &y[i+10], y6 );
        __stfpd( &y[i+12], y7 );
        __stfpd( &y[i+14], y8 );

  }

  for (j = i; j < N; j++)
      y[j] = alpha*x[j];

}


