#include <stdio.h>

/*
 *  Vector element-wise multiplication
 */

double dvvvmul(int n, double *x, double *y, double *z){

  double ans = 0.0;
  for (int i = 0; i < n; ++i)
     ans += x[i]*y[i]*z[i];

  return ans;   
}

double dvv2mul(int n, double *x, double *y){
/* compute dot(x,y.^2); */

  double ans = 0.0;
  for (int i = 0; i < n; ++i)
     ans += x[i]*y[i]*y[i];

  return ans;
}


double dvvvmul_xyz_inc1(int n, double *x, double *y, double *z)
{
  double ans1, ans2, ans3, ans4;
  int i,j,k,l;
  ans1 = ans2 = ans3 = ans4 = 0.0;

  for (i = 0,j=1,k=2,l=3; i < (n-n%4); i+=4,j+=4,k+=4,l+=4){
     ans1 += x[i]*y[i]*z[i];
     ans2 += x[j]*y[j]*z[j];
     ans3 += x[k]*y[k]*z[k];
     ans4 += x[l]*y[l]*z[l]; 
  }

  for (i = (n-n%4); i < n; ++i)
      ans1 += x[i]*y[i]*z[i];       

  ans1 += ans2;
  ans3 += ans4;
  
  return ans3+ans1;
}

double dvvvmul_xyy_inc1(int n, double *x, double *y)
{
  double ans1, ans2, ans3, ans4;
  int i,j,k,l;
  ans1 = ans2 = ans3 = ans4 = 0.0;

  for (i = 0,j=1,k=2,l=3; i < (n-n%4); i+=4,j+=4,k+=4,l+=4){
     ans1 += x[i]*y[i]*y[i];
     ans2 += x[j]*y[j]*y[j];
     ans3 += x[k]*y[k]*y[k];
     ans4 += x[l]*y[l]*y[l];
  }

  for (i = (n-n%4); i < n; ++i)
      ans1 += x[i]*y[i]*y[i];

  ans1 += ans2;
  ans3 += ans4;

  return ans3+ans1;
}




/*
 Feb. 24, 2010
 
 Assumptions:
    1. X,Y,Z are 16 byte aligned 
*/

#if 0 

double dvvvmul_xyz_intrinsics( int N, double *X, double *Y, double *Z )
{

    //if ( ( (int) X%16 != 0) || ( (int) Y%16 != 0) || ( (int) Z%16 != 0) )   {
    //    fprintf(stderr,"dvmul_intrinsics arrays are not aligned:  mod(X,16) = %d, mod(Y,16) = %d, mod(Z,16) = %d\n", (int) X%16, (int) Y%16, (int) Z%16); 
    //   return dvvvmul_xyz_inc1(N,X,Y,Z);
    //}

    int i, j, N1;
    double _Complex x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4, x5, y5, z5, x6, y6, z6, x7, y7, z7, x8, y8, z8, a1, a2, a3, a4;
   
    a1 = __cmplx(0,0);
    a2 = __cmplx(0,0);
    a3 = __cmplx(0,0);
    a4 = __cmplx(0,0);

 
    N1 = N - 16;
    for ( i = 0; i <= N1; i += 16 )
    {
        x1 = __lfpd( &X[i] );
        x2 = __lfpd( &X[i+2] );
        x3 = __lfpd( &X[i+4] );
        x4 = __lfpd( &X[i+6] );
        x5 = __lfpd( &X[i+8] );
        x6 = __lfpd( &X[i+10] );
        x7 = __lfpd( &X[i+12] );
        x8 = __lfpd( &X[i+14] );

        y1 = __lfpd( &Y[i] );
        y2 = __lfpd( &Y[i+2] );
        y3 = __lfpd( &Y[i+4] );
        y4 = __lfpd( &Y[i+6] );
        y5 = __lfpd( &Y[i+8] );
        y6 = __lfpd( &Y[i+10] );
        y7 = __lfpd( &Y[i+12] );
        y8 = __lfpd( &Y[i+14] );


        z1 = __fpmul( y1, x1 );
        z2 = __fpmul( y2, x2 );
        z3 = __fpmul( y3, x3 );
        z4 = __fpmul( y4, x4 );
        z5 = __fpmul( y5, x5 );
        z6 = __fpmul( y6, x6 );
        z7 = __fpmul( y7, x7 );
        z8 = __fpmul( y8, x8 );

      
        x1 = __lfpd( &Z[i] );
        x2 = __lfpd( &Z[i+2] );
        x3 = __lfpd( &Z[i+4] );
        x4 = __lfpd( &Z[i+6] );
        x5 = __lfpd( &Z[i+8] );
        x6 = __lfpd( &Z[i+10] );
        x7 = __lfpd( &Z[i+12] );
        x8 = __lfpd( &Z[i+14] );

        a1 = __fpmadd(a1, z1, x1 );
        a2 = __fpmadd(a2, z2, x2 );
        a3 = __fpmadd(a3, z3, x3 );
        a4 = __fpmadd(a4, z4, x4 );
        a1 = __fpmadd(a1, z5, x5 );
        a2 = __fpmadd(a2, z6, x6 );
        a3 = __fpmadd(a3, z7, x7 );
        a4 = __fpmadd(a4, z8, x8 );

    }
    x1 = __fpadd(a1,a2);
    y1 = __fpadd(a3,a4);
 
    N1 = (N & 15 );
    double ans = 0.0;

    z1 = __fpadd(x1,y1);

    for ( j = 0; j < N1; j++ )
        ans += X[j+i] * Y[j+i] * Z[j+i];    

    ans += __creal(z1) + __cimag(z1); 

    return ans;
}

#else

double dvvvmul_xyz_intrinsics( int N, double *X, double *Y, double *Z )
{

    //if ( ( (int) X%16 != 0) || ( (int) Y%16 != 0) || ( (int) Z%16 != 0) )   {
    //    fprintf(stderr,"dvmul_intrinsics arrays are not aligned:  mod(X,16) = %d, mod(Y,16) = %d, mod(Z,16) = %d\n", (int) X%16, (int) Y%16, (int) Z%16);
    //   return dvvvmul_xyz_inc1(N,X,Y,Z);
   // }

    int i, j, N1;
    double _Complex x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4, a1, a2, a3, a4;

    a1 = __cmplx(0,0);
    a2 = __cmplx(0,0);
    a3 = __cmplx(0,0);
    a4 = __cmplx(0,0);


    N1 = N - 8;
    for ( i = 0; i <= N1; i += 8 )
    {
        x1 = __lfpd( &X[i] );
        x2 = __lfpd( &X[i+2] );
        x3 = __lfpd( &X[i+4] );
        x4 = __lfpd( &X[i+6] );

        y1 = __lfpd( &Y[i] );
        y2 = __lfpd( &Y[i+2] );
        y3 = __lfpd( &Y[i+4] );
        y4 = __lfpd( &Y[i+6] );

        z1 = __fpmul( y1, x1 );
        z2 = __fpmul( y2, x2 );
        z3 = __fpmul( y3, x3 );
        z4 = __fpmul( y4, x4 );

        x1 = __lfpd( &Z[i] );
        x2 = __lfpd( &Z[i+2] );
        x3 = __lfpd( &Z[i+4] );
        x4 = __lfpd( &Z[i+6] );

        a1 = __fpmadd(a1, z1, x1 );
        a2 = __fpmadd(a2, z2, x2 );
        a3 = __fpmadd(a3, z3, x3 );
        a4 = __fpmadd(a4, z4, x4 );

    }
    x1 = __fpadd(a1,a2);
    y1 = __fpadd(a3,a4);

    N1 = (N & 7 );
    double ans = 0.0;

    z1 = __fpadd(x1,y1);

    for ( j = 0; j < N1; j++ )
        ans += X[j+i] * Y[j+i] * Z[j+i];

    ans += __creal(z1) + __cimag(z1);

    return ans;
}
#endif


#if 0 

double dvvvmul_xyy_intrinsics( int N, double *X, double *Y)
{

    //if ( ( (int) X%16 != 0) || ( (int) Y%16 != 0) || ( (int) Z%16 != 0) )   {
    //    fprintf(stderr,"dvmul_intrinsics arrays are not aligned:  mod(X,16) = %d, mod(Y,16) = %d, mod(Z,16) = %d\n", (int) X%16, (int) Y%16, (int) Z%16);
    //   return dvvvmul_xyz_inc1(N,X,Y,Z);
    //}

    int i, j, N1;
    double _Complex x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4, x5, y5, z5, x6, y6, z6, x7, y7, z7, x8, y8, z8, a1, a2, a3, a4;

    a1 = __cmplx(0,0);
    a2 = __cmplx(0,0);
    a3 = __cmplx(0,0);
    a4 = __cmplx(0,0);


    N1 = N - 16;
    for ( i = 0; i <= N1; i += 16 )
    {
        x1 = __lfpd( &X[i] );
        x2 = __lfpd( &X[i+2] );
        x3 = __lfpd( &X[i+4] );
        x4 = __lfpd( &X[i+6] );
        x5 = __lfpd( &X[i+8] );
        x6 = __lfpd( &X[i+10] );
        x7 = __lfpd( &X[i+12] );
        x8 = __lfpd( &X[i+14] );

        y1 = __lfpd( &Y[i] );
        y2 = __lfpd( &Y[i+2] );
        y3 = __lfpd( &Y[i+4] );
        y4 = __lfpd( &Y[i+6] );
        y5 = __lfpd( &Y[i+8] );
        y6 = __lfpd( &Y[i+10] );
        y7 = __lfpd( &Y[i+12] );
        y8 = __lfpd( &Y[i+14] );


        z1 = __fpmul( y1, y1 );
        z2 = __fpmul( y2, y2 );
        z3 = __fpmul( y3, y3 );
        z4 = __fpmul( y4, y4 );
        z5 = __fpmul( y5, y5 );
        z6 = __fpmul( y6, y6 );
        z7 = __fpmul( y7, y7 );
        z8 = __fpmul( y8, y8 );


        a1 = __fpmadd(a1, z1, x1 );
        a2 = __fpmadd(a2, z2, x2 );
        a3 = __fpmadd(a3, z3, x3 );
        a4 = __fpmadd(a4, z4, x4 );
        a1 = __fpmadd(a1, z5, x5 );
        a2 = __fpmadd(a2, z6, x6 );
        a3 = __fpmadd(a3, z7, x7 );
        a4 = __fpmadd(a4, z8, x8 );

    }
    x1 = __fpadd(a1,a2);
    y1 = __fpadd(a3,a4);

    N1 = (N & 15 );
    double ans = 0.0;

    z1 = __fpadd(x1,y1);

    for ( j = 0; j < N1; j++ )
        ans += X[j+i] * Y[j+i] * Y[j+i];

    ans += __creal(z1) + __cimag(z1);

    return ans;
}

#else

double dvvvmul_xyy_intrinsics( int N, double *X, double *Y)
{

    //if ( ( (int) X%16 != 0) || ( (int) Y%16 != 0) )  {
    //    fprintf(stderr,"dvmul_intrinsics arrays are not aligned:  mod(X,16) = %d, mod(Y,16) = %d, mod(Z,16) = %d\n", (int) X%16, (int) Y%16, (int) Z%16);
    //   return dvvvmul_xyy_inc1(N,X,Y);
   // }

    int i, j, N1;
    double _Complex x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4, a1, a2, a3, a4;

    a1 = __cmplx(0,0);
    a2 = __cmplx(0,0);
    a3 = __cmplx(0,0);
    a4 = __cmplx(0,0);


    N1 = N - 8;
    for ( i = 0; i <= N1; i += 8 )
    {
        y1 = __lfpd( &Y[i] );
        y2 = __lfpd( &Y[i+2] );
        y3 = __lfpd( &Y[i+4] );
        y4 = __lfpd( &Y[i+6] );

        x1 = __lfpd( &X[i] );
        x2 = __lfpd( &X[i+2] );
        x3 = __lfpd( &X[i+4] );
        x4 = __lfpd( &X[i+6] );

        z1 = __fpmul( y1, y1 );
        z2 = __fpmul( y2, y2 );
        z3 = __fpmul( y3, y3 );
        z4 = __fpmul( y4, y4 );

        a1 = __fpmadd(a1, z1, x1 );
        a2 = __fpmadd(a2, z2, x2 );
        a3 = __fpmadd(a3, z3, x3 );
        a4 = __fpmadd(a4, z4, x4 );

    }
    x1 = __fpadd(a1,a2);
    y1 = __fpadd(a3,a4);

    N1 = (N & 7 );
    double ans = 0.0;

    z1 = __fpadd(x1,y1);

    for ( j = 0; j < N1; j++ )
        ans += X[j+i] * Y[j+i] * Y[j+i];

    ans += __creal(z1) + __cimag(z1);

    return ans;
}
#endif

